"""Tests for TileDB integration with PyTorch Data API."""

import numpy as np
import pytest
import torch

from tiledb.ml.readers.pytorch import PyTorchTileDBDataLoader, PyTorchTileDBDataset

from .utils import (
    ingest_in_tiledb,
    parametrize_for_dataset,
    rand_array,
    validate_tensor_generator,
)


class TestPyTorchTileDBDataset:
    @parametrize_for_dataset(batch_size=[0], shuffle_buffer_size=[0], num_workers=[0])
    def test_dataset(
        self,
        tmpdir,
        num_rows,
        num_workers,
        x_sparse,
        y_sparse,
        x_shape,
        y_shape,
        num_attrs,
        pass_attrs,
        buffer_bytes,
        batch_size,
        shuffle_buffer_size,
    ):
        with ingest_in_tiledb(
            tmpdir,
            x_data=rand_array(num_rows, *x_shape, sparse=x_sparse),
            y_data=rand_array(num_rows, *y_shape, sparse=y_sparse),
            x_sparse=x_sparse,
            y_sparse=y_sparse,
            num_attrs=num_attrs,
            pass_attrs=pass_attrs,
        ) as kwargs:
            dataset = PyTorchTileDBDataset(buffer_bytes=buffer_bytes, **kwargs)
            assert isinstance(dataset, torch.utils.data.IterableDataset)
            validate_tensor_generator(
                dataset, num_attrs, x_sparse, y_sparse, x_shape, y_shape
            )

    @parametrize_for_dataset()
    def test_dataloader(
        self,
        tmpdir,
        num_rows,
        num_workers,
        x_sparse,
        y_sparse,
        x_shape,
        y_shape,
        num_attrs,
        pass_attrs,
        buffer_bytes,
        batch_size,
        shuffle_buffer_size,
    ):
        if num_workers and (x_sparse or y_sparse):
            pytest.skip("multiple workers not supported with sparse arrays")

        with ingest_in_tiledb(
            tmpdir,
            x_data=rand_array(num_rows, *x_shape, sparse=x_sparse),
            y_data=rand_array(num_rows, *y_shape, sparse=y_sparse),
            x_sparse=x_sparse,
            y_sparse=y_sparse,
            num_attrs=num_attrs,
            pass_attrs=pass_attrs,
        ) as kwargs:
            dataloader = PyTorchTileDBDataLoader(
                num_workers=num_workers,
                buffer_bytes=buffer_bytes,
                batch_size=batch_size,
                shuffle_buffer_size=shuffle_buffer_size,
                **kwargs
            )
            assert isinstance(dataloader, torch.utils.data.DataLoader)
            validate_tensor_generator(
                dataloader, num_attrs, x_sparse, y_sparse, x_shape, y_shape, batch_size
            )

            unique_x_tensors = []
            unique_y_tensors = []
            for i, (x_tensors, y_tensors) in enumerate(dataloader):
                # Keep unique X tensors
                for x_tensor in x_tensors if num_attrs > 1 else [x_tensors]:
                    if x_sparse:
                        x_tensor = x_tensor.to_dense()
                    if not any(torch.equal(x_tensor, t) for t in unique_x_tensors):
                        unique_x_tensors.append(x_tensor)

                # Keep unique Y tensors
                for y_tensor in y_tensors if num_attrs > 1 else [y_tensors]:
                    if y_sparse:
                        y_tensor = y_tensor.to_dense()
                    if not any(torch.equal(y_tensor, t) for t in unique_y_tensors):
                        unique_y_tensors.append(y_tensor)

                assert len(unique_x_tensors) - 1 == i
                assert len(unique_y_tensors) - 1 == i

    @parametrize_for_dataset()
    def test_unequal_num_rows(
        self,
        tmpdir,
        num_rows,
        num_workers,
        x_sparse,
        y_sparse,
        x_shape,
        y_shape,
        num_attrs,
        pass_attrs,
        buffer_bytes,
        batch_size,
        shuffle_buffer_size,
    ):
        with ingest_in_tiledb(
            tmpdir,
            # Add one extra row on X
            x_data=rand_array(num_rows + 1, *x_shape, sparse=x_sparse),
            y_data=rand_array(num_rows, *y_shape, sparse=y_sparse),
            x_sparse=x_sparse,
            y_sparse=y_sparse,
            num_attrs=num_attrs,
            pass_attrs=pass_attrs,
        ) as kwargs:
            with pytest.raises(ValueError) as ex:
                PyTorchTileDBDataLoader(
                    num_workers=num_workers,
                    buffer_bytes=buffer_bytes,
                    batch_size=batch_size,
                    shuffle_buffer_size=shuffle_buffer_size,
                    **kwargs
                )
            assert "X and Y arrays must have the same number of rows" in str(ex.value)

    @parametrize_for_dataset(x_sparse=[True], shuffle_buffer_size=[0], num_workers=[0])
    @pytest.mark.parametrize("csr", [True, False])
    def test_sparse_read_order(
        self,
        tmpdir,
        num_rows,
        num_workers,
        x_sparse,
        y_sparse,
        x_shape,
        y_shape,
        num_attrs,
        pass_attrs,
        buffer_bytes,
        batch_size,
        shuffle_buffer_size,
        csr,
    ):
        x_data = rand_array(num_rows, *x_shape, sparse=x_sparse)
        with ingest_in_tiledb(
            tmpdir,
            x_data=x_data,
            y_data=rand_array(num_rows, *y_shape, sparse=y_sparse),
            x_sparse=x_sparse,
            y_sparse=y_sparse,
            num_attrs=num_attrs,
            pass_attrs=pass_attrs,
        ) as kwargs:
            dataloader = PyTorchTileDBDataLoader(
                buffer_bytes=buffer_bytes,
                batch_size=batch_size,
                shuffle_buffer_size=shuffle_buffer_size,
                num_workers=num_workers,
                csr=csr,
                **kwargs
            )
            generated_x_data = np.concatenate(
                [
                    (x if num_attrs == 1 else x[0]).to_dense().numpy()
                    for x, y in dataloader
                ]
            )
            np.testing.assert_array_almost_equal(generated_x_data, x_data)



"""Functionality for saving and loading Sklearn models as TileDB arrays"""

import pickle
from typing import Optional

import numpy as np
import sklearn
from sklearn import config_context
from sklearn.base import BaseEstimator

import tiledb

from ._cloud_utils import update_file_properties
from .base import Meta, TileDBModel, Timestamp, current_milli_time


class SklearnTileDBModel(TileDBModel[BaseEstimator]):
    """
    Class that implements all functionality needed to save Sklearn models as
    TileDB arrays and load Sklearn models from TileDB arrays.
    """

    Framework = "SKLEARN"
    FrameworkVersion = sklearn.__version__

    def save(self, *, update: bool = False, meta: Optional[Meta] = None) -> None:
        """
        Save a Sklearn model as a TileDB array.

        :param update: Whether we should update any existing TileDB array model at the
            target location.
        :param meta: Extra metadata to save in a TileDB array.
        """
        # Serialize model
        serialized_model = self._serialize_model()

        # Create TileDB model array
        if not update:
            self._create_array()

        self._write_array(serialized_model=serialized_model, meta=meta)

    def load(self, *, timestamp: Optional[Timestamp] = None) -> BaseEstimator:
        """
        Load a Sklearn model from a TileDB array.

        :param timestamp: Range of timestamps to load fragments of the array which live
            in the specified time range.
        :return: A Sklearn model object.
        """
        # TODO: Change timestamp when issue in core is resolved

        model_array = tiledb.open(self.uri, ctx=self.ctx, timestamp=timestamp)
        model_array_results = model_array[:]
        model = pickle.loads(model_array_results["model_params"].item(0))
        return model

    def preview(self, *, display: str = "text") -> str:
        """
        Create a text representation of the model.

        :param display. If ‘diagram’, estimators will be displayed as a diagram in an
            HTML format when shown in a jupyter notebook. If ‘text’, estimators will be
            displayed as text.
        :return. A string representation of the models internal configuration.
        """
        if self.model:
            with config_context(display=display):
                return str(self.model)
        else:
            return ""

    def _create_array(self) -> None:
        """Create a TileDB array for a Sklearn model."""
        dom = tiledb.Domain(
            tiledb.Dim(
                name="model", domain=(1, 1), tile=1, dtype=np.int32, ctx=self.ctx
            ),
        )

        attrs = [
            tiledb.Attr(
                name="model_params",
                dtype=bytes,
                var=True,
                filters=tiledb.FilterList([tiledb.ZstdFilter()]),
                ctx=self.ctx,
            ),
        ]

        schema = tiledb.ArraySchema(domain=dom, sparse=False, attrs=attrs, ctx=self.ctx)

        tiledb.Array.create(self.uri, schema, ctx=self.ctx)

        # In case we are on TileDB-Cloud we have to update model array's file properties
        if self.namespace:
            update_file_properties(self.uri, self._file_properties)

    def _write_array(self, serialized_model: bytes, meta: Optional[Meta]) -> None:
        """
        Write a Sklearn model to a TileDB array.

        :param serialized_model: A pickled sklearn model.
        :param meta: Extra metadata to save in a TileDB array.
        """
        # TODO: Change timestamp when issue in core is resolved

        with tiledb.open(
            self.uri, "w", timestamp=current_milli_time(), ctx=self.ctx
        ) as tf_model_tiledb:
            # Insertion in TileDB array
            tf_model_tiledb[:] = {"model_params": np.array([serialized_model])}
            self.update_model_metadata(array=tf_model_tiledb, meta=meta)

    def _serialize_model(self) -> bytes:
        """
        Serialize a Sklearn model with pickle.

        :return: Pickled Sklearn model.
        """
        return pickle.dumps(self.model, protocol=4)



"""Tests for TileDB PyTorch model save and load."""

import glob
import inspect
import os
import pickle
import platform
import shutil

import pytest
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optimizers
from torch.utils.tensorboard import SummaryWriter

import tiledb
from tiledb.ml.models.pytorch import PyTorchTileDBModel


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        # 1 input image channel, 6 output channels, 3x3 square convolution
        # kernel
        self.conv1 = nn.Conv2d(1, 6, 3)
        self.conv2 = nn.Conv2d(6, 16, 3)
        # an affine operation: y = Wx + b
        self.fc1 = nn.Linear(16 * 6 * 6, 120)  # 6*6 from image dimension
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        # Max pooling over a (2, 2) window
        x = F.max_pool2d(F.relu(self.conv1(x)), (2, 2))
        # If the size is a square you can only specify a single number
        x = F.max_pool2d(F.relu(self.conv2(x)), 2)
        x = x.view(-1, self.num_flat_features(x))
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

    def num_flat_features(self, x):
        size = x.size()[1:]  # all dimensions except the batch dimension
        num_features = 1
        for s in size:
            num_features *= s
        return num_features


class SeqNeuralNetwork(nn.Module):
    def __init__(self):
        super(SeqNeuralNetwork, self).__init__()
        self.flatten = nn.Flatten()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(14 * 14, 256),
            nn.ReLU(),
            nn.Linear(256, 256),
            nn.ReLU(),
            nn.Linear(256, 10),
            nn.ReLU(),
        )

    def forward(self, x):
        x = self.flatten(x)
        logits = self.linear_relu_stack(x)
        return logits


class ConvNet(nn.Module):
    def __init__(self):
        super(ConvNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


net = pytest.mark.parametrize("net", [ConvNet, Net, SeqNeuralNetwork])


class TestPyTorchModel:
    @net
    @pytest.mark.parametrize(
        "optimizer",
        [
            getattr(optimizers, name)
            for name, obj in inspect.getmembers(optimizers)
            if inspect.isclass(obj) and name != "Optimizer"
        ],
    )
    def test_save(self, tmpdir, net, optimizer):
        EPOCH = 5
        LOSS = 0.4
        model = net()
        saved_optimizer = optimizer(model.parameters(), lr=0.001)
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = PyTorchTileDBModel(
            uri=tiledb_array, model=model, optimizer=saved_optimizer
        )

        tiledb_obj.save(
            model_info={
                "epoch": EPOCH,
                "loss": LOSS,
            }
        )
        loaded_net = net()
        loaded_optimizer = optimizer(loaded_net.parameters(), lr=0.001)
        model_out_dict = tiledb_obj.load(model=loaded_net, optimizer=loaded_optimizer)

        assert model_out_dict["epoch"] == EPOCH
        assert model_out_dict["loss"] == LOSS

        # Check model parameters
        for key_item_1, key_item_2 in zip(
            model.state_dict().items(), loaded_net.state_dict().items()
        ):
            assert torch.equal(key_item_1[1], key_item_2[1])

        # Check optimizer parameters
        for key_item_1, key_item_2 in zip(
            saved_optimizer.state_dict().items(), loaded_optimizer.state_dict().items()
        ):
            assert all([a == b for a, b in zip(key_item_1[1], key_item_2[1])])

    @net
    def test_preview(self, tmpdir, net):
        # With model given as argument
        model = net()
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = PyTorchTileDBModel(uri=tiledb_array, model=model)
        assert type(tiledb_obj.preview()) == str
        tiledb_obj_none = PyTorchTileDBModel(uri=tiledb_array, model=None)
        assert tiledb_obj_none.preview() == ""

    @net
    def test_file_properties(self, tmpdir, net):
        model = net()
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = PyTorchTileDBModel(uri=tiledb_array, model=model)

        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK"] == "PYTORCH"
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_STAGE"] == "STAGING"
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_PYTHON_VERSION"]
            == platform.python_version()
        )
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION"]
            == torch.__version__
        )
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_PREVIEW"] == str(model)

    @net
    def test_tensorboard_callback_meta(self, tmpdir, net):
        model = net()
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = PyTorchTileDBModel(uri=tiledb_array, model=model)

        # SummaryWriter creates file(s) under log_dir
        log_dir = os.path.join(tmpdir, "logs")
        writer = SummaryWriter(log_dir=log_dir)
        log_files = read_files(log_dir)
        assert log_files

        tiledb_obj.save(update=False, summary_writer=writer)
        with tiledb.open(tiledb_array) as A:
            assert pickle.loads(A.meta["__TENSORBOARD__"]) == log_files
        shutil.rmtree(log_dir)

        # Loading the event data should create local files
        tiledb_obj.load_tensorboard()
        new_log_files = read_files(log_dir)
        assert new_log_files == log_files

        custom_dir = os.path.join(tmpdir, "custom_log")
        tiledb_obj.load_tensorboard(target_dir=custom_dir)
        new_log_files = read_files(custom_dir)
        assert len(new_log_files) == len(log_files)
        for new_file, old_file in zip(new_log_files.values(), log_files.values()):
            assert new_file == old_file


def read_files(dirpath):
    files = {}
    for path in glob.glob(f"{dirpath}/*"):
        with open(path, "rb") as f:
            files[path] = f.read()
    return files


class TestPyTorchModelCloud:
    def test_get_cloud_uri_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = Net()
        uri = os.path.join(tmpdir, "model_array")

        mock_get_cloud_uri = mocker.patch(
            "tiledb.ml.models.base.get_cloud_uri", return_value=uri
        )

        _ = PyTorchTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_cloud_uri.assert_called_once_with(uri, "test_namespace")

    def test_get_s3_prefix_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = Net()
        uri = os.path.join(tmpdir, "model_array")

        mock_get_s3_prefix = mocker.patch(
            "tiledb.ml.models._cloud_utils.get_s3_prefix", return_value="s3 prefix"
        )

        _ = PyTorchTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_s3_prefix.assert_called_once_with("test_namespace")

    def test_update_file_properties_call(self, tmpdir, mocker):
        model = Net()
        uri = os.path.join(tmpdir, "model_array")

        mocker.patch("tiledb.ml.models.base.get_cloud_uri", return_value=uri)

        tiledb_obj = PyTorchTileDBModel(
            uri=uri, namespace="test_namespace", model=model
        )

        mock_update_file_properties = mocker.patch(
            "tiledb.ml.models.pytorch.update_file_properties", return_value=None
        )
        mocker.patch("tiledb.ml.models.pytorch.PyTorchTileDBModel._write_array")

        tiledb_obj.save()

        file_properties_dict = {
            "TILEDB_ML_MODEL_ML_FRAMEWORK": "PYTORCH",
            "TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION": torch.__version__,
            "TILEDB_ML_MODEL_STAGE": "STAGING",
            "TILEDB_ML_MODEL_PYTHON_VERSION": platform.python_version(),
            "TILEDB_ML_MODEL_PREVIEW": str(model),
        }

        mock_update_file_properties.assert_called_once_with(uri, file_properties_dict)

    def test_exception_raise_file_property_in_meta_error(self, tmpdir):
        model = Net()
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = PyTorchTileDBModel(uri=tiledb_array, model=model)
        with pytest.raises(ValueError) as ex:
            tiledb_obj.save(
                meta={"TILEDB_ML_MODEL_ML_FRAMEWORK": "TILEDB_ML_MODEL_ML_FRAMEWORK"},
            )

        assert "Please avoid using file property key names as metadata keys!" in str(
            ex.value
        )



"""Tests for TileDB Tensorflow Keras model save and load."""

import io
import os
import pickle
import platform
import shutil

import numpy as np
import pytest
import tensorflow as tf

import tiledb
from tiledb.ml.models.tensorflow_keras import (
    TensorflowKerasTileDBModel,
    tf_keras_is_keras,
)

if tf_keras_is_keras:
    from keras import testing_utils
else:
    from tensorflow.python.keras import testing_utils


# Suppress all Tensorflow messages
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

batch_get_value = tf.keras.backend.batch_get_value


class ConfigSubclassModel(tf.keras.Model):
    def __init__(self, hidden_units):
        super(ConfigSubclassModel, self).__init__()
        self.hidden_units = hidden_units
        self.dense_layers = [tf.keras.layers.Dense(u) for u in hidden_units]

    def call(self, inputs):
        x = inputs
        for layer in self.dense_layers:
            x = layer(x)
        return x

    def get_config(self):
        return {"hidden_units": self.hidden_units}

    @classmethod
    def from_config(cls, config):
        return cls(**config)


def test_load_tiledb_error_with_wrong_uri():
    tiledb_model_obj = TensorflowKerasTileDBModel(uri="dummy_uri")
    with pytest.raises(tiledb.TileDBError):
        tiledb_model_obj.load(compile_model=False)


loss_optimizer_metrics = pytest.mark.parametrize(
    "loss,optimizer,metrics",
    [
        (None, None, None),
        ("binary_crossentropy", "rmsprop", "accuracy"),
        ("mean_squared_error", "rmsprop", "categorical_accuracy"),
        ("sparse_categorical_crossentropy", "sgd", "sparse_categorical_crossentropy"),
    ],
)

api = pytest.mark.parametrize(
    "api",
    [
        testing_utils.get_small_sequential_mlp,
        testing_utils.get_small_functional_mlp,
        ConfigSubclassModel,
    ],
)


class TestTensorflowKerasModel:
    @api
    @loss_optimizer_metrics
    def test_save_model_to_tiledb_array(self, tmpdir, api, loss, optimizer, metrics):
        model = (
            api(num_hidden=1, num_classes=2, input_dim=3)
            if api != ConfigSubclassModel
            else api(hidden_units=[16, 16, 10])
        )

        tiledb_uri = os.path.join(tmpdir, "model_array")

        # Compiles the model if optimizer is present
        if optimizer:
            model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        if not model.built:
            model.build(tuple(np.random.randint(20, size=2)))
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True if optimizer else False)
        assert tiledb.array_exists(tiledb_uri)

    @api
    @loss_optimizer_metrics
    def test_save_model_to_tiledb_array_predictions(
        self, tmpdir, api, loss, optimizer, metrics
    ):
        model = (
            api(num_hidden=1, num_classes=2, input_dim=3)
            if api != ConfigSubclassModel
            else api(hidden_units=[16, 16, 10])
        )

        tiledb_uri = os.path.join(tmpdir, "model_array")

        # Compiles the model if optimizer is present
        if optimizer:
            model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        input_shape = tuple(np.random.randint(20, size=2))
        if not model.built:
            model.build(input_shape)
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True if optimizer else False)

        loaded_model = (
            tiledb_model_obj.load(
                compile_model=False,
                custom_objects={"ConfigSubclassModel": ConfigSubclassModel},
                input_shape=input_shape,
            )
            if api == ConfigSubclassModel
            else tiledb_model_obj.load(compile_model=False)
        )

        data = np.random.rand(100, input_shape[-1] if api == ConfigSubclassModel else 3)

        # Assert model predictions are equal
        np.testing.assert_array_equal(loaded_model.predict(data), model.predict(data))

    @api
    @loss_optimizer_metrics
    def test_save_model_to_tiledb_array_weights(
        self, tmpdir, api, loss, optimizer, metrics
    ):
        model = (
            api(num_hidden=1, num_classes=2, input_dim=3)
            if api != ConfigSubclassModel
            else api(hidden_units=[16, 16, 10])
        )

        tiledb_uri = os.path.join(tmpdir, "model_array")

        # Compiles the model if optimizer is present
        if optimizer:
            model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        input_shape = tuple(np.random.randint(20, size=2))
        if not model.built:
            model.build(input_shape)
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)

        tiledb_model_obj.save(include_optimizer=True if optimizer else False)
        loaded_model = (
            tiledb_model_obj.load(compile_model=True if optimizer else False)
            if api != ConfigSubclassModel
            else tiledb_model_obj.load(
                compile_model=True if optimizer else False,
                custom_objects={"ConfigSubclassModel": ConfigSubclassModel},
                input_shape=input_shape,
            )
        )

        data = np.random.rand(100, input_shape[-1] if api == ConfigSubclassModel else 3)

        if optimizer:
            model_opt_weights = batch_get_value(model.optimizer.weights)
            loaded_opt_weights = batch_get_value(loaded_model.optimizer.weights)

            # Assert optimizer weights are equal
            for weight_model, weight_loaded_model in zip(
                model_opt_weights, loaded_opt_weights
            ):
                np.testing.assert_array_equal(weight_model, weight_loaded_model)

            # Assert model predictions are equal
            np.testing.assert_array_equal(
                loaded_model.predict(data), model.predict(data)
            )

    @loss_optimizer_metrics
    def test_save_load_with_dense_features(self, tmpdir, loss, optimizer, metrics):
        if optimizer is None:
            pytest.skip()
        cols = [
            tf.feature_column.numeric_column("a"),
            tf.feature_column.indicator_column(
                tf.feature_column.categorical_column_with_vocabulary_list(
                    "b", ["one", "two"]
                )
            ),
        ]
        input_layers = {
            "a": tf.keras.layers.Input(shape=(1,), name="a"),
            "b": tf.keras.layers.Input(shape=(1,), name="b", dtype="string"),
        }

        fc_layer = tf.keras.layers.DenseFeatures(cols)(input_layers)
        output = tf.keras.layers.Dense(10)(fc_layer)

        model = tf.keras.Model(input_layers, output)

        model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True)
        loaded_model = tiledb_model_obj.load(compile_model=True)

        model_opt_weights = batch_get_value(model.optimizer.weights)
        loaded_opt_weights = batch_get_value(loaded_model.optimizer.weights)

        # Assert optimizer weights are equal
        for weight_model, weight_loaded_model in zip(
            model_opt_weights, loaded_opt_weights
        ):
            np.testing.assert_array_equal(weight_model, weight_loaded_model)

        inputs_a = np.arange(10).reshape(10, 1)
        inputs_b = np.arange(10).reshape(10, 1).astype("str")

        # Assert model predictions are equal
        np.testing.assert_array_equal(
            loaded_model.predict({"a": inputs_a, "b": inputs_b}),
            model.predict({"a": inputs_a, "b": inputs_b}),
        )

    @loss_optimizer_metrics
    def test_save_load_with_sequence_features(self, tmpdir, loss, optimizer, metrics):
        if optimizer is None:
            pytest.skip()

        cols = [
            tf.feature_column.sequence_numeric_column("a"),
            tf.feature_column.indicator_column(
                tf.feature_column.sequence_categorical_column_with_vocabulary_list(
                    "b", ["one", "two"]
                )
            ),
        ]
        input_layers = {
            "a": tf.keras.layers.Input(shape=(None, 1), sparse=True, name="a"),
            "b": tf.keras.layers.Input(
                shape=(None, 1), sparse=True, name="b", dtype="string"
            ),
        }

        fc_layer, _ = tf.keras.experimental.SequenceFeatures(cols)(input_layers)
        x = tf.keras.layers.GRU(32)(fc_layer)
        output = tf.keras.layers.Dense(10)(x)

        model = tf.keras.Model(input_layers, output)

        model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True)
        loaded_model = tiledb_model_obj.load(compile_model=True)

        model_opt_weights = batch_get_value(model.optimizer.weights)
        loaded_opt_weights = batch_get_value(loaded_model.optimizer.weights)

        # Assert optimizer weights are equal
        for weight_model, weight_loaded_model in zip(
            model_opt_weights, loaded_opt_weights
        ):
            np.testing.assert_array_equal(weight_model, weight_loaded_model)

        batch_size = 10
        timesteps = 1

        values_a = np.arange(10, dtype=np.float32)
        indices_a = np.zeros((10, 3), dtype=np.int64)
        indices_a[:, 0] = np.arange(10)
        inputs_a = tf.SparseTensor(indices_a, values_a, (batch_size, timesteps, 1))

        values_b = np.zeros(10, dtype=np.str)
        indices_b = np.zeros((10, 3), dtype=np.int64)
        indices_b[:, 0] = np.arange(10)
        inputs_b = tf.SparseTensor(indices_b, values_b, (batch_size, timesteps, 1))

        # Assert model predictions are equal
        np.testing.assert_array_equal(
            loaded_model.predict({"a": inputs_a, "b": inputs_b}, steps=1),
            model.predict({"a": inputs_a, "b": inputs_b}, steps=1),
        )

    def test_functional_model_save_load_with_custom_loss_and_metric(self, tmpdir):
        inputs = tf.keras.layers.Input(shape=(4,))
        x = tf.keras.layers.Dense(8, activation="relu")(inputs)
        outputs = tf.keras.layers.Dense(3, activation="softmax")(x)
        model = tf.keras.Model(inputs=inputs, outputs=outputs)
        custom_loss = tf.keras.layers.Lambda(lambda x: tf.keras.backend.sum(x * x))(x)
        model.add_loss(custom_loss)
        model.add_metric(custom_loss, aggregation="mean", name="custom_loss")

        model.compile(
            loss="sparse_categorical_crossentropy",
            optimizer="sgd",
            metrics=["sparse_categorical_crossentropy"],
        )

        data_x = np.random.normal(size=(32, 4))
        data_y = np.random.randint(0, 3, size=32)
        model.train_on_batch(data_x, data_y)

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True)
        loaded_model = tiledb_model_obj.load(compile_model=True)

        # Assert all evaluation results are the same.
        assert all(
            [
                a == pytest.approx(b, 1e-9)
                for a, b in zip(
                    model.evaluate(data_x, data_y),
                    loaded_model.evaluate(data_x, data_y),
                )
            ]
        )

        # Assert model predictions are equal
        np.testing.assert_array_equal(
            loaded_model.predict(data_x), model.predict(data_x)
        )

    def test_save_load_for_rnn_layers(self, tmpdir):
        inputs = tf.keras.layers.Input([10, 10], name="train_input")
        rnn_layers = [
            tf.keras.layers.LSTMCell(size, recurrent_dropout=0, name="rnn_cell%d" % i)
            for i, size in enumerate([32, 32])
        ]
        rnn_output = tf.keras.layers.RNN(
            rnn_layers, return_sequences=True, name="rnn_layer"
        )(inputs)
        pred_feat = tf.keras.layers.Dense(10, name="prediction_features")(rnn_output)
        pred = tf.keras.layers.Softmax()(pred_feat)
        model = tf.keras.Model(inputs=[inputs], outputs=[pred, pred_feat])

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=False)
        loaded_model = tiledb_model_obj.load(compile_model=False)

        data = np.random.rand(50, 10, 10)

        # Assert model predictions are equal
        np.testing.assert_array_equal(loaded_model.predict(data), model.predict(data))

    def test_sequential_model_save_load_without_input_shape(self, tmpdir):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Dense(2))
        model.add(tf.keras.layers.RepeatVector(3))
        model.add(tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(3)))
        model.compile(
            loss="mean_squared_error",
            optimizer="rmsprop",
            metrics="categorical_accuracy",
            weighted_metrics="categorical_accuracy",
            sample_weight_mode="temporal",
        )
        data_x = np.random.random((1, 3))
        data_y = np.random.random((1, 3, 3))
        model.train_on_batch(data_x, data_y)

        tiledb_uri = os.path.join(tmpdir, "model_array")
        tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
        tiledb_model_obj.save(include_optimizer=True)
        loaded_model = tiledb_model_obj.load(compile_model=True)

        # Assert model predictions are equal
        np.testing.assert_array_equal(
            loaded_model.predict(data_x), model.predict(data_x)
        )

    @api
    @loss_optimizer_metrics
    def test_preview(self, tmpdir, api, loss, optimizer, metrics):
        model = (
            api(num_hidden=1, num_classes=2, input_dim=3)
            if api != ConfigSubclassModel
            else api(hidden_units=[16, 16, 10])
        )

        tiledb_uri = os.path.join(tmpdir, "model_array")

        # Compiles the model if optimizer is present
        if optimizer:
            model.compile(loss=loss, optimizer=optimizer, metrics=[metrics])

        # With model given as argument
        if model.built:
            tiledb_model_obj = TensorflowKerasTileDBModel(uri=tiledb_uri, model=model)
            s = io.StringIO()
            model.summary(print_fn=lambda x: s.write(x + "\n"))
            model_summary = s.getvalue()
            assert tiledb_model_obj.preview() == model_summary
        else:
            # Model should be built before preview it
            with pytest.raises(ValueError):
                tiledb_model_obj = TensorflowKerasTileDBModel(
                    uri=tiledb_uri, model=model
                )
                tiledb_model_obj.preview()

        # When model is None then preview returns empty string
        tiledb_model_obj_none = TensorflowKerasTileDBModel(uri=tiledb_uri, model=None)
        assert tiledb_model_obj_none.preview() == ""


class TestTensorflowKerasModelCloud:
    def test_file_properties(self, tmpdir):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))

        # Get model summary in a string
        s = io.StringIO()
        model.summary(print_fn=lambda x: s.write(x + "\n"))
        model_summary = s.getvalue()

        uri = os.path.join(tmpdir, "model_array")
        tiledb_obj = TensorflowKerasTileDBModel(uri=uri, model=model)

        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK"]
            == "TENSORFLOW KERAS"
        )
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_STAGE"] == "STAGING"
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_PYTHON_VERSION"]
            == platform.python_version()
        )
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION"]
            == tf.__version__
        )
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_PREVIEW"] == model_summary

    def test_get_cloud_uri_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))
        uri = os.path.join(tmpdir, "model_array")

        mock_get_cloud_uri = mocker.patch(
            "tiledb.ml.models.base.get_cloud_uri", return_value=uri
        )

        _ = TensorflowKerasTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_cloud_uri.assert_called_once_with(uri, "test_namespace")

    def test_get_s3_prefix_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))
        uri = os.path.join(tmpdir, "model_array")

        mock_get_s3_prefix = mocker.patch(
            "tiledb.ml.models._cloud_utils.get_s3_prefix", return_value="s3 prefix"
        )

        _ = TensorflowKerasTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_s3_prefix.assert_called_once_with("test_namespace")

    def test_update_file_properties_call(self, tmpdir, mocker):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))

        # Get model summary in a string
        s = io.StringIO()
        model.summary(print_fn=lambda x: s.write(x + "\n"))
        model_summary = s.getvalue()

        uri = os.path.join(tmpdir, "model_array")

        mocker.patch("tiledb.ml.models.base.get_cloud_uri", return_value=uri)

        tiledb_obj = TensorflowKerasTileDBModel(
            uri=uri, namespace="test_namespace", model=model
        )

        mock_update_file_properties = mocker.patch(
            "tiledb.ml.models.tensorflow_keras.update_file_properties",
            return_value=None,
        )
        mocker.patch(
            "tiledb.ml.models.tensorflow_keras.TensorflowKerasTileDBModel._write_array"
        )

        tiledb_obj.save()

        file_properties_dict = {
            "TILEDB_ML_MODEL_ML_FRAMEWORK": "TENSORFLOW KERAS",
            "TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION": tf.__version__,
            "TILEDB_ML_MODEL_STAGE": "STAGING",
            "TILEDB_ML_MODEL_PYTHON_VERSION": platform.python_version(),
            "TILEDB_ML_MODEL_PREVIEW": model_summary,
        }

        mock_update_file_properties.assert_called_once_with(uri, file_properties_dict)

    def test_exception_raise_file_property_in_meta_error(self, tmpdir):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = TensorflowKerasTileDBModel(uri=tiledb_array, model=model)
        with pytest.raises(ValueError) as ex:
            tiledb_obj.save(
                meta={"TILEDB_ML_MODEL_ML_FRAMEWORK": "TILEDB_ML_MODEL_ML_FRAMEWORK"},
            )

        assert "Please avoid using file property key names as metadata keys!" in str(
            ex.value
        )

    def test_tensorboard_callback_meta(self, tmpdir):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Flatten(input_shape=(10, 10)))
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = TensorflowKerasTileDBModel(uri=tiledb_array, model=model)

        cb = [tf.keras.callbacks.TensorBoard(log_dir=tmpdir)]

        os.mkdir(os.path.join(tmpdir, "train"))
        with open(os.path.join(tmpdir, "train", "foo_tfevents_1"), "wb") as f:
            f.write(b"test_bytes_1")
        with open(os.path.join(tmpdir, "train", "bar_tfevents_2"), "wb") as f:
            f.write(b"test_bytes_2")

        tiledb_obj.save(include_callbacks=cb)
        with tiledb.open(tiledb_array) as A:
            assert pickle.loads(A.meta["__TENSORBOARD__"]) == {
                os.path.join(tmpdir, "train", "foo_tfevents_1"): b"test_bytes_1",
                os.path.join(tmpdir, "train", "bar_tfevents_2"): b"test_bytes_2",
            }
        shutil.rmtree(os.path.join(tmpdir, "train"))

        # Loading the event data should create local files
        tiledb_obj.load_tensorboard()
        with open(os.path.join(tmpdir, "train", "foo_tfevents_1"), "rb") as f:
            assert f.read() == b"test_bytes_1"
        with open(os.path.join(tmpdir, "train", "bar_tfevents_2"), "rb") as f:
            assert f.read() == b"test_bytes_2"

        custom_dir = os.path.join(tmpdir, "custom_log")
        tiledb_obj.load_tensorboard(target_dir=custom_dir)
        with open(os.path.join(custom_dir, "foo_tfevents_1"), "rb") as f:
            assert f.read() == b"test_bytes_1"
        with open(os.path.join(custom_dir, "bar_tfevents_2"), "rb") as f:
            assert f.read() == b"test_bytes_2"



"""Functionality for loading data from TileDB arrays to the Tensorflow Data API."""

import math
from typing import Optional, Sequence, Tuple, Union

import numpy as np
import sparse
import tensorflow as tf

import tiledb

from ._buffer_utils import get_attr_names, get_buffer_size
from ._tensor_gen import (
    TileDBNumpyGenerator,
    TileDBSparseGenerator,
    TileDBTensorGenerator,
    iter_slices,
)


def TensorflowTileDBDataset(
    x_array: tiledb.Array,
    y_array: tiledb.Array,
    *,
    batch_size: int,
    buffer_bytes: Optional[int] = None,
    shuffle_buffer_size: int = 0,
    prefetch: int = tf.data.AUTOTUNE,
    x_attrs: Sequence[str] = (),
    y_attrs: Sequence[str] = (),
    num_workers: int = 0,
) -> tf.data.Dataset:
    """Return a tf.data.Dataset for loading data from TileDB arrays.

    :param x_array: TileDB array of the features.
    :param y_array: TileDB array of the labels.
    :param batch_size: Size of each batch.
    :param buffer_bytes: Maximum size (in bytes) of memory to allocate for reading from
        each array (default=`tiledb.default_ctx().config()["sm.memory_budget"]`).
    :param shuffle_buffer_size: Number of elements from which this dataset will sample.
    :param prefetch: Maximum number of batches that will be buffered when prefetching.
        By default, the buffer size is dynamically tuned.
    :param x_attrs: Attribute names of x_array.
    :param y_attrs: Attribute names of y_array.
    :param num_workers: If greater than zero, create a threadpool of `num_workers` threads
        used to fetch inputs asynchronously and in parallel. Note: yielded batches may
        be shuffled even if `shuffle_buffer_size` is zero when `num_workers` > 1.
    """
    # Check that x_array and y_array have the same number of rows
    rows: int = x_array.shape[0]
    if rows != y_array.shape[0]:
        raise ValueError("X and Y arrays must have the same number of rows")

    if not x_attrs:
        x_attrs = get_attr_names(x_array.schema)
    if not y_attrs:
        y_attrs = get_attr_names(y_array.schema)

    x_gen = _get_tensor_generator(x_array, x_attrs)
    y_gen = _get_tensor_generator(y_array, y_attrs)
    x_buffer_size = get_buffer_size(x_array, x_attrs, buffer_bytes)
    y_buffer_size = get_buffer_size(y_array, y_attrs, buffer_bytes)

    def bounded_dataset(bounds: Union[Tuple[int, int], tf.Tensor]) -> tf.data.Dataset:
        x_dataset = tf.data.Dataset.from_generator(
            x_gen.iter_tensors,
            args=(x_buffer_size, bounds[0], bounds[1]),
            output_signature=_get_tensor_specs(x_array.schema, x_attrs),
        )
        y_dataset = tf.data.Dataset.from_generator(
            y_gen.iter_tensors,
            args=(y_buffer_size, bounds[0], bounds[1]),
            output_signature=_get_tensor_specs(y_array.schema, y_attrs),
        )
        return tf.data.Dataset.zip((x_dataset.unbatch(), y_dataset.unbatch()))

    if num_workers:
        per_worker = int(math.ceil(rows / num_workers))
        offsets = [(s.start, s.stop) for s in iter_slices(0, rows, per_worker)]
        offsets_tensor = tf.convert_to_tensor(offsets, dtype=tf.int64)
        offsets_dataset = tf.data.Dataset.from_tensor_slices(offsets_tensor)
        dataset = offsets_dataset.interleave(
            bounded_dataset, num_parallel_calls=num_workers, deterministic=False
        )
    else:
        dataset = bounded_dataset((0, rows))

    if shuffle_buffer_size > 0:
        dataset = dataset.shuffle(shuffle_buffer_size)
    return dataset.batch(batch_size).prefetch(prefetch)


TensorSpec = Union[tf.TensorSpec, tf.SparseTensorSpec]


def _get_tensor_specs(
    schema: tiledb.ArraySchema, attrs: Sequence[str]
) -> Union[TensorSpec, Sequence[TensorSpec]]:
    cls = tf.SparseTensorSpec if schema.sparse else tf.TensorSpec
    specs = (
        cls(shape=(None, *schema.shape[1:]), dtype=schema.attr(attr).dtype)
        for attr in attrs
    )
    return tuple(specs) if len(attrs) > 1 else next(specs)


def _get_tensor_generator(
    array: tiledb.Array, attrs: Sequence[str]
) -> TileDBTensorGenerator[Union[np.ndarray, tf.SparseTensor]]:
    if not array.schema.sparse:
        return TileDBNumpyGenerator(array, attrs)
    else:
        return TileDBSparseGenerator(array, attrs, _coo_to_sparse_tensor)


def _coo_to_sparse_tensor(coo: sparse.COO) -> tf.SparseTensor:
    return tf.SparseTensor(coo.coords.T, coo.data, coo.shape)



"""Tests for TileDB Sklearn model save and load."""

import os
import platform
from itertools import zip_longest

import pytest
import sklearn
import sklearn.base

from tiledb.ml.models.sklearn import SklearnTileDBModel


def iter_models(*modules):
    for estim_name, estim_class in sklearn.utils.all_estimators():
        if str(estim_class).split(".")[1] in modules:
            yield estim_class


@pytest.mark.parametrize(
    "net",
    list(iter_models("svm", "linear_model", "naive_bayes", "tree")),
)
class TestSklearnModel:
    def test_save_load(self, tmpdir, net):
        tiledb_array = os.path.join(tmpdir, "test_array")
        model = net()
        tiledb_sklearn_obj = SklearnTileDBModel(uri=tiledb_array, model=model)
        tiledb_sklearn_obj.save()
        loaded_model = tiledb_sklearn_obj.load()
        assert all(
            [
                a == b
                for a, b in zip_longest(model.get_params(), loaded_model.get_params())
            ]
        )

    def test_preview(self, tmpdir, net):
        # With model as argument
        tiledb_array = os.path.join(tmpdir, "test_array")
        model = net()
        tiledb_sklearn_obj = SklearnTileDBModel(uri=tiledb_array, model=model)
        assert type(tiledb_sklearn_obj.preview()) == str
        tiledb_sklearn_obj_none = SklearnTileDBModel(uri=tiledb_array, model=None)
        assert tiledb_sklearn_obj_none.preview() == ""

    def test_file_properties(self, tmpdir, net):
        model = net()
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = SklearnTileDBModel(uri=tiledb_array, model=model)
        tiledb_obj.save()

        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK"] == "SKLEARN"
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_STAGE"] == "STAGING"
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_PYTHON_VERSION"]
            == platform.python_version()
        )
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION"]
            == sklearn.__version__
        )
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_PREVIEW"] == str(model)


class TestSklearnModelCloud:
    def test_get_cloud_uri_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = sklearn.linear_model.LinearRegression()
        uri = os.path.join(tmpdir, "model_array")

        mock_get_cloud_uri = mocker.patch(
            "tiledb.ml.models.base.get_cloud_uri", return_value=uri
        )

        _ = SklearnTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_cloud_uri.assert_called_once_with(uri, "test_namespace")

    def test_get_s3_prefix_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = sklearn.linear_model.LinearRegression()
        uri = os.path.join(tmpdir, "model_array")

        mock_get_s3_prefix = mocker.patch(
            "tiledb.ml.models._cloud_utils.get_s3_prefix", return_value="s3 prefix"
        )

        _ = SklearnTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_s3_prefix.assert_called_once_with("test_namespace")

    def test_update_file_properties_call(self, tmpdir, mocker):
        model = sklearn.linear_model.LinearRegression()
        uri = os.path.join(tmpdir, "model_array")

        mocker.patch("tiledb.ml.models.base.get_cloud_uri", return_value=uri)

        tiledb_obj = SklearnTileDBModel(
            uri=uri, namespace="test_namespace", model=model
        )

        mock_update_file_properties = mocker.patch(
            "tiledb.ml.models.sklearn.update_file_properties", return_value=None
        )
        mocker.patch("tiledb.ml.models.sklearn.SklearnTileDBModel._write_array")

        tiledb_obj.save()

        file_properties_dict = {
            "TILEDB_ML_MODEL_ML_FRAMEWORK": "SKLEARN",
            "TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION": sklearn.__version__,
            "TILEDB_ML_MODEL_STAGE": "STAGING",
            "TILEDB_ML_MODEL_PYTHON_VERSION": platform.python_version(),
            "TILEDB_ML_MODEL_PREVIEW": str(model),
        }

        mock_update_file_properties.assert_called_once_with(uri, file_properties_dict)

    def test_exception_raise_file_property_in_meta_error(self, tmpdir):
        model = sklearn.linear_model.LinearRegression()
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = SklearnTileDBModel(uri=tiledb_array, model=model)
        with pytest.raises(ValueError) as ex:
            tiledb_obj.save(
                meta={"TILEDB_ML_MODEL_ML_FRAMEWORK": "TILEDB_ML_MODEL_ML_FRAMEWORK"},
            )

        assert "Please avoid using file property key names as metadata keys!" in str(
            ex.value
        )



"""Functionality for loading data from TileDB arrays to the PyTorch Dataloader API."""

import itertools as it
import math
import operator
import random
from typing import (
    Callable,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
import scipy.sparse
import sparse
import torch

try:
    # torch>=1.10
    sparse_csr_tensor = torch.sparse_csr_tensor
except AttributeError:
    # torch=1.9
    sparse_csr_tensor = torch._sparse_csr_tensor

import tiledb

from ._buffer_utils import get_attr_names, get_buffer_size
from ._tensor_gen import (
    TileDBNumpyGenerator,
    TileDBSparseGenerator,
    TileDBTensorGenerator,
)

TensorLike = Union[np.ndarray, sparse.COO, scipy.sparse.csr_matrix]
TensorLikeOrSequence = Union[TensorLike, Sequence[TensorLike]]
XY = Tuple[TensorLikeOrSequence, TensorLikeOrSequence]


def PyTorchTileDBDataLoader(
    x_array: tiledb.Array,
    y_array: tiledb.Array,
    *,
    batch_size: int,
    buffer_bytes: Optional[int] = None,
    shuffle_buffer_size: int = 0,
    prefetch: int = 2,
    x_attrs: Sequence[str] = (),
    y_attrs: Sequence[str] = (),
    num_workers: int = 0,
    csr: bool = True,
) -> torch.utils.data.DataLoader:
    """Return a DataLoader for loading data from TileDB arrays.

    :param x_array: TileDB array of the features.
    :param y_array: TileDB array of the labels.
    :param batch_size: Size of each batch.
    :param buffer_bytes: Maximum size (in bytes) of memory to allocate for reading
        from each array (default=`tiledb.default_ctx().config()["sm.memory_budget"]`).
    :param shuffle_buffer_size: Number of elements from which this dataset will sample.
    :param prefetch: Number of samples loaded in advance by each worker. Not applicable
        (and should not be given) when `num_workers` is 0.
    :param x_attrs: Attribute names of x_array.
    :param y_attrs: Attribute names of y_array.
    :param num_workers: how many subprocesses to use for data loading. 0 means that the
        data will be loaded in the main process. Note: yielded batches may be shuffled
        even if `shuffle_buffer_size` is zero when `num_workers` > 1.
    :param csr: For sparse 2D arrays, whether to return CSR tensors instead of COO.
    """
    if not x_attrs:
        x_attrs = get_attr_names(x_array.schema)
    if not y_attrs:
        y_attrs = get_attr_names(y_array.schema)

    return torch.utils.data.DataLoader(
        dataset=PyTorchTileDBDataset(
            x_array, y_array, buffer_bytes, shuffle_buffer_size, x_attrs, y_attrs
        ),
        batch_size=batch_size,
        prefetch_factor=prefetch,
        num_workers=num_workers,
        collate_fn=_CompositeCollator(
            _get_tensor_collator(x_array, csr, len(x_attrs)),
            _get_tensor_collator(y_array, csr, len(y_attrs)),
        ),
    )


class PyTorchTileDBDataset(torch.utils.data.IterableDataset[XY]):
    def __init__(
        self,
        x_array: tiledb.Array,
        y_array: tiledb.Array,
        buffer_bytes: Optional[int] = None,
        shuffle_buffer_size: int = 0,
        x_attrs: Sequence[str] = (),
        y_attrs: Sequence[str] = (),
    ):
        super().__init__()
        rows: int = x_array.shape[0]
        if rows != y_array.shape[0]:
            raise ValueError("X and Y arrays must have the same number of rows")

        if not x_attrs:
            x_attrs = get_attr_names(x_array.schema)
        if not y_attrs:
            y_attrs = get_attr_names(y_array.schema)

        self._x_gen = _get_tensor_generator(x_array, x_attrs)
        self._y_gen = _get_tensor_generator(y_array, y_attrs)
        self._x_buffer_size = get_buffer_size(x_array, x_attrs, buffer_bytes)
        self._y_buffer_size = get_buffer_size(y_array, y_attrs, buffer_bytes)
        self._rows = rows
        self._shuffle_buffer_size = shuffle_buffer_size

    def __iter__(self) -> Iterator[XY]:
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is not None:
            for gen in self._x_gen, self._y_gen:
                if isinstance(gen, TileDBSparseGenerator):
                    raise NotImplementedError(
                        "https://github.com/pytorch/pytorch/issues/20248"
                    )
            per_worker = int(math.ceil(self._rows / worker_info.num_workers))
            start = worker_info.id * per_worker
            stop = min(start + per_worker, self._rows)
        else:
            start = 0
            stop = self._rows

        def iter_rows(
            gen: TileDBTensorGenerator[TensorLike], buffer_size: int
        ) -> Iterator[TensorLikeOrSequence]:
            iter_tensors = gen.iter_tensors(buffer_size, start, stop)
            if gen.single_attr:
                return (row for tensor in iter_tensors for row in tensor)
            else:
                return (row for tensors in iter_tensors for row in zip(*tensors))

        rows: Iterator[XY] = zip(
            iter_rows(self._x_gen, self._x_buffer_size),
            iter_rows(self._y_gen, self._y_buffer_size),
        )
        if self._shuffle_buffer_size > 0:
            rows = _iter_shuffled(rows, self._shuffle_buffer_size)
        return rows


def _get_tensor_generator(
    array: tiledb.Array, attrs: Sequence[str]
) -> TileDBTensorGenerator[TensorLike]:
    if not array.schema.sparse:
        return TileDBNumpyGenerator(array, attrs)
    elif array.ndim == 2:
        return TileDBSparseGenerator(array, attrs, operator.methodcaller("tocsr"))
    else:
        return TileDBSparseGenerator(array, attrs, lambda x: x)


_SingleCollator = Callable[[Sequence[TensorLike]], torch.Tensor]


class _CompositeCollator:
    """
    A callable for collating "rows" of data by a separate collator for each "column".
    Returns the collated columns collected into a tuple.
    """

    def __init__(self, *collators: _SingleCollator):
        self._collators = collators

    def __call__(self, rows: Sequence[Sequence[TensorLike]]) -> Sequence[torch.Tensor]:
        columns = tuple(zip(*rows))
        collators = self._collators
        assert len(columns) == len(collators)
        return tuple(collator(column) for collator, column in zip(collators, columns))


def _ndarray_collate(arrays: Sequence[np.ndarray]) -> torch.Tensor:
    """Collate multiple Numpy arrays to a torch.Tensor with strided layout."""
    # Specialized version of default_collate for collating Numpy arrays
    # Faster than `torch.as_tensor(arrays)` (https://github.com/pytorch/pytorch/pull/51731)
    # and `torch.stack([torch.as_tensor(array) for array in arrays]])`
    return torch.from_numpy(np.stack(arrays))


def _sparse_coo_collate(arrays: Sequence[sparse.COO]) -> torch.Tensor:
    """Collate multiple sparse.COO arrays to a torch.Tensor with sparse_coo layout."""
    stacked = sparse.stack(arrays)
    return torch.sparse_coo_tensor(stacked.coords, stacked.data, stacked.shape)


def _csr_to_coo_collate(arrays: Sequence[scipy.sparse.csr_matrix]) -> torch.Tensor:
    """Collate multiple Scipy CSR matrices to a torch.Tensor with sparse_coo layout."""
    stacked = scipy.sparse.vstack(arrays).tocoo()
    coords = np.stack((stacked.row, stacked.col))
    return torch.sparse_coo_tensor(coords, stacked.data, stacked.shape)


def _csr_collate(arrays: Sequence[scipy.sparse.csr_matrix]) -> torch.Tensor:
    """Collate multiple Scipy CSR matrices to a torch.Tensor with sparse_csr layout."""
    stacked = scipy.sparse.vstack(arrays)
    return sparse_csr_tensor(
        torch.from_numpy(stacked.indptr),
        torch.from_numpy(stacked.indices),
        stacked.data,
        stacked.shape,
    )


def _get_tensor_collator(
    array: tiledb.Array, csr: bool, num_attrs: int
) -> Union[_SingleCollator, _CompositeCollator]:
    if not array.schema.sparse:
        collator = _ndarray_collate
    elif array.ndim != 2:
        collator = _sparse_coo_collate
    elif csr:
        collator = _csr_collate
    else:
        collator = _csr_to_coo_collate

    if num_attrs == 1:
        return collator
    else:
        return _CompositeCollator(*it.repeat(collator, num_attrs))


_T = TypeVar("_T")


def _iter_shuffled(iterable: Iterable[_T], buffer_size: int) -> Iterator[_T]:
    """
    Shuffle the given iterable with a buffer.

    The buffer with `buffer_size` is filled with elements from the iterable first.
    Then, each item will be yielded from the buffer by reservoir sampling via iterator.

    """
    iterator = iter(iterable)
    buffer = list(it.islice(iterator, buffer_size))
    randrange = random.randrange
    for x in iterator:
        idx = randrange(0, buffer_size)
        yield buffer[idx]
        buffer[idx] = x
    random.shuffle(buffer)
    while buffer:
        yield buffer.pop()



from math import ceil
from typing import Optional, Sequence

import numpy as np

import tiledb


def get_buffer_size(
    array: tiledb.Array,
    attrs: Sequence[str] = (),
    memory_budget: Optional[int] = None,
    start_offset: int = 0,
    stop_offset: int = 0,
) -> int:
    if not stop_offset:
        stop_offset = array.shape[0]
    if array.schema.sparse:
        # TODO: implement get_max_buffer_size() for sparse arrays
        row_bytes = estimate_row_bytes(array, attrs, start_offset, stop_offset)
        buffer_size = (memory_budget or 100 * 1024**2) // row_bytes
    else:
        buffer_size = get_max_buffer_size(array.schema, attrs, memory_budget)
    # clip the buffer size between 1 and total number of rows
    return max(1, min(buffer_size, stop_offset - start_offset))


def estimate_row_bytes(
    array: tiledb.Array,
    attrs: Sequence[str] = (),
    start_offset: int = 0,
    stop_offset: int = 0,
) -> int:
    """
    Estimate the size in bytes of a TileDB array row.

    A "row" is a slice with the first dimension fixed.
    - For dense arrays, each row consists of a fixed number of cells. The size of each
      cell depends on the given attributes (or all array attributes by default).
    - For sparse arrays, each row consists of a variable number of non-empty cells. The
      size of each non-empty cell depends on all dimension coordinates as well as the
      given attributes (or all array attributes by default).
    """
    schema = array.schema
    if not attrs:
        attrs = get_attr_names(schema)

    if not schema.sparse:
        # for dense arrays the size of each row is fixed and can be computed exactly
        row_cells = np.prod(schema.shape[1:])
        cell_bytes = sum(schema.attr(attr).dtype.itemsize for attr in attrs)
        est_row_bytes = row_cells * cell_bytes
    else:
        # for sparse arrays the size of each row is variable and can only be estimated
        if not stop_offset:
            stop_offset = schema.shape[0]
        query = array.query(return_incomplete=True)
        # .multi_index[] is inclusive, so we need to subtract 1 to stop_offset
        indexer = query.multi_index[start_offset : stop_offset - 1]
        est_rs = indexer.estimated_result_sizes()
        dims = get_dim_names(schema)
        est_total_bytes = sum(est_rs[key].data_bytes for key in (*dims, *attrs))
        est_row_bytes = est_total_bytes / (stop_offset - start_offset)
    return int(est_row_bytes)


def get_max_buffer_size(
    schema: tiledb.ArraySchema,
    attrs: Sequence[str] = (),
    memory_budget: Optional[int] = None,
) -> int:
    """
    Get the maximum number of "rows" that can be read from an array with the given schema
    without incurring incomplete reads.

    A "row" is a slice with the first dimension fixed.

    :param schema: The array schema.
    :param attrs: The attributes to read; defaults to all array attributes.
    :param memory_budget: The maximum amount of memory to use. This is bounded by
        `tiledb.default_ctx().config()["sm.memory_budget"]`, which is also used as the
        default memory_budget.
    """
    if schema.sparse:
        raise NotImplementedError(
            "get_max_buffer_size() is not implemented for sparse arrays"
        )

    config_memory_budget = int(tiledb.default_ctx().config()["sm.memory_budget"])
    if memory_budget is None or memory_budget > config_memory_budget:
        memory_budget = config_memory_budget

    # The memory budget should be large enough to read the cells of the largest attribute
    if not attrs:
        attrs = get_attr_names(schema)
    bytes_per_cell = max(schema.attr(attr).dtype.itemsize for attr in attrs)

    # We want to be reading tiles following the tile extents along each dimension.
    # The number of cells for each such tile is the product of all tile extents.
    dim_tiles = tuple(int(schema.domain.dim(idx).tile) for idx in range(schema.ndim))
    cells_per_tile = np.prod(dim_tiles)

    # Reading a slice of dim_tiles[0] rows requires reading a number of tiles that
    # depends on the size and tile extent of each dimension after the first one.
    assert len(schema.shape) == len(dim_tiles)
    tiles_per_slice = np.prod(
        tuple(
            ceil(dim_size / dim_tile)
            for dim_size, dim_tile in zip(schema.shape[1:], dim_tiles[1:])
        )
    )

    # Compute the size in bytes of each slice of dim_tiles[0] rows
    bytes_per_slice = int(bytes_per_cell * cells_per_tile * tiles_per_slice)

    # Compute the number of slices that fit within the memory budget
    num_slices = memory_budget // bytes_per_slice

    # Compute the total number of rows to slice
    return dim_tiles[0] * num_slices


def get_attr_names(schema: tiledb.ArraySchema) -> Sequence[str]:
    return tuple(schema.attr(idx).name for idx in range(schema.nattr))


def get_dim_names(schema: tiledb.ArraySchema) -> Sequence[str]:
    return tuple(schema.domain.dim(idx).name for idx in range(schema.ndim))



"""Base Class for saving and loading machine learning models."""

import platform
import time
from abc import ABC, abstractmethod
from enum import Enum, unique
from typing import Any, Generic, Mapping, Optional, Tuple, TypeVar

import tiledb

from ._cloud_utils import get_cloud_uri

Model = TypeVar("Model")
Meta = Mapping[str, Any]
Timestamp = Tuple[int, int]


@unique
class ModelFileProperties(Enum):
    """
    Enum Class that contains all model array file properties.
    """

    TILEDB_ML_MODEL_ML_FRAMEWORK = "TILEDB_ML_MODEL_ML_FRAMEWORK"
    TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION = "TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION"
    TILEDB_ML_MODEL_STAGE = "TILEDB_ML_MODEL_STAGE"
    TILEDB_ML_MODEL_PYTHON_VERSION = "TILEDB_ML_MODEL_PYTHON_VERSION"
    TILEDB_ML_MODEL_PREVIEW = "TILEDB_ML_MODEL_PREVIEW"


def current_milli_time() -> int:
    return round(time.time() * 1000)


class TileDBModel(ABC, Generic[Model]):
    """
    This is the base class for all TileDB model storage functionalities, i.e,
    store machine learning models (Tensorflow, PyTorch, etc) as TileDB arrays.
    """

    Framework: str
    FrameworkVersion: str

    def __init__(
        self,
        uri: str,
        namespace: Optional[str] = None,
        ctx: Optional[tiledb.Ctx] = None,
        model: Optional[Model] = None,
    ):
        """
        Base class for saving machine learning models as TileDB arrays
        and loading machine learning models from TileDB arrays. In case we need to interact
        with TileDB-Cloud we have to pass user's TileDB-Cloud namespace. If we don't
        models will be saved locally.

        :param uri: TileDB array uri
        :param namespace: In case we want to interact (save, load, update, check) with
            models on TileDB-Cloud we need the user's namespace on TileDB-Cloud.
            Moreover, array's uri must have an s3 prefix.
        :param ctx: TileDB Context.
        :param model: Machine learning model based on the framework we are using.
        """
        self.namespace = namespace
        self.ctx = ctx
        self.model = model
        self.uri = get_cloud_uri(uri, namespace) if namespace else uri
        self._file_properties = {
            ModelFileProperties.TILEDB_ML_MODEL_ML_FRAMEWORK.value: self.Framework,
            ModelFileProperties.TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION.value: self.FrameworkVersion,
            ModelFileProperties.TILEDB_ML_MODEL_STAGE.value: "STAGING",
            ModelFileProperties.TILEDB_ML_MODEL_PYTHON_VERSION.value: platform.python_version(),
            ModelFileProperties.TILEDB_ML_MODEL_PREVIEW.value: self.preview(),
        }

    @abstractmethod
    def save(self, *, update: bool = False, meta: Optional[Meta] = None) -> None:
        """Abstract method that saves a machine learning model as a TileDB array.

        :param update: Whether we should update any existing TileDB array model at the
            target location.
        :param meta: Extra metadata to save in a TileDB array.
        """

    @abstractmethod
    def load(self, *, timestamp: Optional[Timestamp] = None) -> Model:
        """Abstract method that loads a machine learning model from a TileDB array.

        :param timestamp: Range of timestamps to load fragments of the array which live
            in the specified time range.
        """

    @abstractmethod
    def preview(self) -> str:
        """Abstract method that previews a machine learning model."""

    def update_model_metadata(
        self, array: tiledb.Array, meta: Optional[Meta] = None
    ) -> None:
        """
        Update the metadata in a TileDB model array. File properties also go in the metadata section.

        :param array: A TileDB model array.
        :param meta: A mapping with the <key, value> pairs to be inserted in array's metadata.
        """
        # Raise ValueError in case users provide metadata with the same keys as file properties.
        if meta:
            if not meta.keys().isdisjoint(self._file_properties.keys()):
                raise ValueError(
                    "Please avoid using file property key names as metadata keys!"
                )
            for key, value in meta.items():
                array.meta[key] = value
        for key, value in self._file_properties.items():
            array.meta[key] = value



import itertools as it
import os
import uuid
from contextlib import contextmanager

import numpy as np
import pytest
import scipy.sparse
import sparse
import tensorflow as tf
import torch

import tiledb


def parametrize_for_dataset(
    num_rows=(107,),
    num_workers=(0, 2),
    x_sparse=(True, False),
    y_sparse=(True, False),
    x_shape=((10,), (10, 3)),
    y_shape=((5,), (5, 2)),
    num_attrs=(1, 2),
    pass_attrs=(True, False),
    batch_size=(8,),
    buffer_bytes=(1024, None),
    shuffle_buffer_size=(0, 16),
):
    argnames = [
        "num_rows",
        "num_workers",
        "x_sparse",
        "y_sparse",
        "x_shape",
        "y_shape",
        "num_attrs",
        "pass_attrs",
        "buffer_bytes",
        "batch_size",
        "shuffle_buffer_size",
    ]
    argvalues = it.product(
        num_rows,
        num_workers,
        x_sparse,
        y_sparse,
        x_shape,
        y_shape,
        num_attrs,
        pass_attrs,
        buffer_bytes,
        batch_size,
        shuffle_buffer_size,
    )
    return pytest.mark.parametrize(argnames, argvalues)


@contextmanager
def ingest_in_tiledb(
    tmpdir,
    x_data,
    y_data,
    x_sparse,
    y_sparse,
    num_attrs,
    pass_attrs,
):
    """Context manager for ingest data into TileDB.

    Yield the keyword arguments for instantiating a TiledbDataset.
    """
    array_uuid = str(uuid.uuid4())
    x_uri = os.path.join(tmpdir, "x_" + array_uuid)
    y_uri = os.path.join(tmpdir, "y_" + array_uuid)
    _ingest_in_tiledb(x_uri, x_data, x_sparse, num_attrs)
    _ingest_in_tiledb(y_uri, y_data, y_sparse, num_attrs)
    attrs = [f"features_{attr}" for attr in range(num_attrs)] if pass_attrs else []
    with tiledb.open(x_uri) as x_array, tiledb.open(y_uri) as y_array:
        yield dict(x_array=x_array, y_array=y_array, x_attrs=attrs, y_attrs=attrs)


def _ingest_in_tiledb(uri: str, data: np.ndarray, sparse: bool, num_attrs: int) -> None:
    dims = [
        tiledb.Dim(
            name=f"dim_{dim}",
            domain=(0, data.shape[dim] - 1),
            tile=np.random.randint(1, data.shape[dim] + 1),
            dtype=np.int32,
        )
        for dim in range(data.ndim)
    ]

    # TileDB schema
    schema = tiledb.ArraySchema(
        domain=tiledb.Domain(*dims),
        sparse=sparse,
        attrs=[
            tiledb.Attr(name=f"features_{attr}", dtype=np.float32)
            for attr in range(num_attrs)
        ],
    )

    # Create the (empty) array on disk.
    tiledb.Array.create(uri, schema)

    # Ingest
    with tiledb.open(uri, "w") as tiledb_array:
        idx = np.nonzero(data) if sparse else slice(None)
        tiledb_array[idx] = {f"features_{attr}": data[idx] for attr in range(num_attrs)}


def rand_array(num_rows: int, *row_shape: int, sparse: bool = False) -> np.ndarray:
    """Create a random array of shape (num_rows, *row_shape).

    :param num_rows: Number of rows of the array (i.e. first dimension size).
    :param row_shape: Shape of each row (i.e. remaining dimension sizes).
    :param sparse:
      - If false, all values will be in the (0, 1) range.
      - If true, only `num_rows` values will be in the (0, 1) range, the rest will be 0.
        Note: some rows may be all zeros.
    """
    shape = (num_rows, *row_shape)
    if sparse:
        a = np.zeros(shape)
        flat_idxs = np.random.choice(a.size, size=len(a), replace=False)
        a.flat[flat_idxs] = np.random.random(len(flat_idxs))
    else:
        a = np.random.random(shape)
    assert a.shape == shape
    return a


def validate_tensor_generator(
    generator, num_attrs, x_sparse, y_sparse, x_shape, y_shape, batch_size=None
):
    for x_tensors, y_tensors in generator:
        for x_tensor in x_tensors if num_attrs > 1 else [x_tensors]:
            _validate_tensor(x_tensor, x_sparse, x_shape, batch_size)
        for y_tensor in y_tensors if num_attrs > 1 else [y_tensors]:
            _validate_tensor(y_tensor, y_sparse, y_shape, batch_size)


def _validate_tensor(tensor, expected_sparse, expected_shape, batch_size=None):
    if batch_size is None and not isinstance(tensor, scipy.sparse.spmatrix):
        row_shape = tensor.shape
    else:
        num_rows, *row_shape = tensor.shape
        if batch_size is None:
            assert isinstance(tensor, scipy.sparse.spmatrix)
            assert num_rows == 1
        else:
            # num_rows may be less than batch_size
            assert num_rows <= batch_size, (num_rows, batch_size)
    assert tuple(row_shape) == expected_shape
    assert _is_sparse(tensor) == expected_sparse


def _is_sparse(tensor):
    if isinstance(tensor, torch.Tensor):
        return tensor.layout in (torch.sparse_coo, torch.sparse_csr)
    if isinstance(tensor, (scipy.sparse.spmatrix, sparse.SparseArray, tf.SparseTensor)):
        return True
    if isinstance(tensor, (np.ndarray, tf.Tensor)):
        return False
    assert False, f"Unknown tensor type: {type(tensor)}"



"""Tests for TileDB integration with Tensorflow Data API."""

import numpy as np
import pytest
import tensorflow as tf

from tiledb.ml.readers.tensorflow import TensorflowTileDBDataset

from .utils import (
    ingest_in_tiledb,
    parametrize_for_dataset,
    rand_array,
    validate_tensor_generator,
)


class TestTensorflowTileDBDataset:
    @parametrize_for_dataset()
    def test_dataset(
        self,
        tmpdir,
        num_rows,
        num_workers,
        x_sparse,
        y_sparse,
        x_shape,
        y_shape,
        num_attrs,
        pass_attrs,
        buffer_bytes,
        batch_size,
        shuffle_buffer_size,
    ):
        with ingest_in_tiledb(
            tmpdir,
            x_data=rand_array(num_rows, *x_shape, sparse=x_sparse),
            y_data=rand_array(num_rows, *y_shape, sparse=y_sparse),
            x_sparse=x_sparse,
            y_sparse=y_sparse,
            num_attrs=num_attrs,
            pass_attrs=pass_attrs,
        ) as kwargs:
            dataset = TensorflowTileDBDataset(
                buffer_bytes=buffer_bytes,
                batch_size=batch_size,
                shuffle_buffer_size=shuffle_buffer_size,
                num_workers=num_workers,
                **kwargs,
            )
            assert isinstance(dataset, tf.data.Dataset)
            validate_tensor_generator(
                dataset, num_attrs, x_sparse, y_sparse, x_shape, y_shape, batch_size
            )

    @parametrize_for_dataset()
    def test_unequal_num_rows(
        self,
        tmpdir,
        num_rows,
        num_workers,
        x_sparse,
        y_sparse,
        x_shape,
        y_shape,
        num_attrs,
        pass_attrs,
        buffer_bytes,
        batch_size,
        shuffle_buffer_size,
    ):
        with ingest_in_tiledb(
            tmpdir,
            # Add one extra row on X
            x_data=rand_array(num_rows + 1, *x_shape, sparse=x_sparse),
            y_data=rand_array(num_rows, *y_shape, sparse=y_sparse),
            x_sparse=x_sparse,
            y_sparse=y_sparse,
            num_attrs=num_attrs,
            pass_attrs=pass_attrs,
        ) as kwargs:
            with pytest.raises(ValueError) as ex:
                TensorflowTileDBDataset(
                    buffer_bytes=buffer_bytes,
                    batch_size=batch_size,
                    shuffle_buffer_size=shuffle_buffer_size,
                    num_workers=num_workers,
                    **kwargs,
                )
            assert "X and Y arrays must have the same number of rows" in str(ex.value)

    @parametrize_for_dataset(x_sparse=[True], shuffle_buffer_size=[0], num_workers=[0])
    def test_sparse_read_order(
        self,
        tmpdir,
        num_rows,
        num_workers,
        x_sparse,
        y_sparse,
        x_shape,
        y_shape,
        num_attrs,
        pass_attrs,
        buffer_bytes,
        batch_size,
        shuffle_buffer_size,
    ):
        x_data = rand_array(num_rows, *x_shape, sparse=x_sparse)
        with ingest_in_tiledb(
            tmpdir,
            x_data=x_data,
            y_data=rand_array(num_rows, *y_shape, sparse=y_sparse),
            x_sparse=x_sparse,
            y_sparse=y_sparse,
            num_attrs=num_attrs,
            pass_attrs=pass_attrs,
        ) as kwargs:
            dataset = TensorflowTileDBDataset(
                buffer_bytes=buffer_bytes,
                batch_size=batch_size,
                shuffle_buffer_size=shuffle_buffer_size,
                num_workers=num_workers,
                **kwargs,
            )
            generated_x_data = np.concatenate(
                [
                    tf.sparse.to_dense(
                        tf.sparse.reorder(x_tensors if num_attrs == 1 else x_tensors[0])
                    )
                    for x_tensors, y_tensors in dataset
                ]
            )
            np.testing.assert_array_almost_equal(generated_x_data, x_data)



import setuptools

tensorflow = ["tensorflow>=2.5"]
pytorch = ["torch>=1.9"]
sklearn = ["scikit-learn>=1.0"]
cloud = ["tiledb-cloud"]
full = sorted({"torchvision", *tensorflow, *pytorch, *sklearn, *cloud})

setuptools.setup(
    setup_requires=["setuptools_scm"],
    use_scm_version={
        "version_scheme": "guess-next-dev",
        "local_scheme": "dirty-tag",
        "write_to": "tiledb/ml/version.py",
    },
    install_requires=["sparse", "tiledb>=0.14"],
    extras_require={
        "tensorflow": tensorflow,
        "pytorch": pytorch,
        "sklearn": sklearn,
        "cloud": cloud,
        "full": full,
    },
)



from abc import ABC, abstractmethod
from operator import itemgetter
from typing import Callable, Generic, Iterator, Sequence, TypeVar, Union

import numpy as np
import sparse

import tiledb

T = TypeVar("T")


class TileDBTensorGenerator(ABC, Generic[T]):
    """Generator of tensors read from a TileDB array."""

    def __init__(self, array: tiledb.Array, attrs: Sequence[str]) -> None:
        """
        :param array: TileDB array to read from.
        :param attrs: Attribute names of array to read.
        """
        self._array = array
        self._attrs = attrs

    @property
    def single_attr(self) -> bool:
        """Whether this generator reads a single attribute."""
        return len(self._attrs) == 1

    @abstractmethod
    def iter_tensors(
        self, buffer_size: int, start_offset: int, stop_offset: int
    ) -> Union[Iterator[T], Iterator[Sequence[T]]]:
        """
        Generate batches of tensors.

        Each yielded batch is either:
        - a sequence of N tensors where `N == len(self.attrs)` if N > 1, or
        - a single tensor if N == 1.
        Each tensor is a `T` instance of shape `(buffer_size, *self.array.shape[1:])`.

        :param buffer_size: Size of each slice of rows to read.
        :param start_offset: Start row offset; defaults to 0.
        :param stop_offset: Stop row offset; defaults to number of rows.
        """


class TileDBNumpyGenerator(TileDBTensorGenerator[np.ndarray]):
    """Generator of Numpy arrays read from a TileDB array."""

    def iter_tensors(
        self, buffer_size: int, start_offset: int, stop_offset: int
    ) -> Union[Iterator[np.ndarray], Iterator[Sequence[np.ndarray]]]:
        query = self._array.query(attrs=self._attrs)
        read_slices = iter_slices(start_offset, stop_offset, buffer_size)
        buffers = map(query.__getitem__, read_slices)
        return map(itemgetter(*self._attrs), buffers)


class TileDBSparseGenerator(TileDBTensorGenerator[T]):
    """Generator of sparse.COO tensors read from a TileDB array."""

    def __init__(
        self,
        array: tiledb.Array,
        attrs: Sequence[str],
        from_coo: Callable[[sparse.COO], T],
    ) -> None:
        super().__init__(array, attrs)
        self._from_coo = from_coo
        self._dims = tuple(array.domain.dim(i).name for i in range(array.ndim))

    def iter_tensors(
        self, buffer_size: int, start_offset: int, stop_offset: int
    ) -> Union[Iterator[T], Iterator[Sequence[T]]]:
        shape = list(self._array.shape)
        query = self._array.query(attrs=self._attrs)
        get_coords = itemgetter(*self._dims)
        get_data = itemgetter(*self._attrs)
        single_attr = self.single_attr
        for read_slice in iter_slices(start_offset, stop_offset, buffer_size):
            shape[0] = read_slice.stop - read_slice.start
            buffer = query[read_slice]
            coords = get_coords(buffer)
            # normalize the first coordinate dimension to start at start_offset
            np.subtract(coords[0], read_slice.start, out=coords[0])
            data = get_data(buffer)
            if single_attr:
                yield self._from_coo(sparse.COO(coords, data, shape))
            else:
                yield tuple(self._from_coo(sparse.COO(coords, d, shape)) for d in data)


def iter_slices(start: int, stop: int, step: int) -> Iterator[slice]:
    offsets = range(start, stop, step)
    yield from map(slice, offsets, offsets[1:])
    yield slice(offsets[-1], stop)



import glob
import os
import pickle
from typing import Mapping, Optional

import tiledb

from .base import Timestamp

_KEY = "__TENSORBOARD__"


def save_tensorboard(log_dir: str) -> Mapping[str, bytes]:
    event_files = {}
    for path in glob.glob(f"{log_dir}/*tfevents*"):
        with open(path, "rb") as f:
            event_files[path] = f.read()
    return {_KEY: pickle.dumps(event_files, protocol=4)}


def load_tensorboard(
    uri: str,
    ctx: Optional[tiledb.Ctx] = None,
    target_dir: Optional[str] = None,
    timestamp: Optional[Timestamp] = None,
) -> None:
    with tiledb.open(uri, ctx=ctx, timestamp=timestamp) as model_array:
        for path, file_bytes in pickle.loads(model_array.meta[_KEY]).items():
            log_dir = target_dir if target_dir else os.path.dirname(path)
            if not os.path.exists(log_dir):
                os.mkdir(log_dir)
            with open(os.path.join(log_dir, os.path.basename(path)), "wb") as f:
                f.write(file_bytes)



import numpy as np
import pytest

import tiledb
from tiledb.ml.readers._buffer_utils import estimate_row_bytes, get_max_buffer_size


@pytest.fixture
def dense_uri(tmp_path):
    uri = str(tmp_path / "dense")
    schema = tiledb.ArraySchema(
        sparse=False,
        domain=tiledb.Domain(
            tiledb.Dim(name="d0", domain=(0, 9999), dtype=np.uint32, tile=123),
            tiledb.Dim(name="d1", domain=(1, 5), dtype=np.uint32, tile=2),
            tiledb.Dim(name="d2", domain=(1, 2), dtype=np.uint32, tile=1),
        ),
        attrs=[
            tiledb.Attr(name="af8", dtype=np.float64),
            tiledb.Attr(name="af4", dtype=np.float32),
            tiledb.Attr(name="au1", dtype=np.uint8),
        ],
    )
    tiledb.Array.create(uri, schema)
    with tiledb.open(uri, "w") as a:
        size = a.schema.domain.size
        a[:] = {
            "af8": np.random.rand(size),
            "af4": np.random.rand(size).astype(np.float32),
            "au1": np.random.randint(128, size=size, dtype=np.uint8),
        }
    return uri


@pytest.fixture
def sparse_uri(tmp_path):
    uri = str(tmp_path / "sparse")
    schema = tiledb.ArraySchema(
        sparse=True,
        allows_duplicates=True,
        domain=tiledb.Domain(
            tiledb.Dim(name="d0", domain=(0, 999), dtype=np.int32),
            tiledb.Dim(name="d1", domain=(-5000, 5000), dtype=np.int32),
            tiledb.Dim(name="d2", domain=(1, 10), dtype=np.int32),
        ),
        attrs=[
            tiledb.Attr(name="af8", dtype=np.float64),
            tiledb.Attr(name="af4", dtype=np.float32),
            tiledb.Attr(name="au1", dtype=np.uint8),
        ],
    )
    tiledb.Array.create(uri, schema)
    with tiledb.open(uri, "w") as a:
        num_rows = 1000
        cells_per_row = 3
        num_cells = num_rows * cells_per_row
        d0 = np.concatenate(
            [np.arange(num_rows, dtype=np.uint32) for _ in range(cells_per_row)]
        )
        d1 = np.random.randint(-5000, 5001, num_cells).astype(np.int32)
        d2 = np.random.randint(1, 11, num_cells).astype(np.uint16)
        a[d0, d1, d2] = {
            "af8": np.random.rand(num_cells),
            "af4": np.random.rand(num_cells).astype(np.float32),
            "au1": np.random.randint(128, size=num_cells, dtype=np.uint8),
        }
    return uri


def test_estimate_row_bytes_dense(dense_uri):
    with tiledb.open(dense_uri) as a:
        # 10 cells/row, 8+4+1=13 bytes/cell
        assert estimate_row_bytes(a) == 130
        # 10 cells/row, 8+1=9 bytes/cell
        assert estimate_row_bytes(a, attrs=["af8", "au1"]) == 90
        # 10 cells/row, 4 bytes/cell
        assert estimate_row_bytes(a, attrs=["af4"]) == 40


def test_estimate_row_bytes_sparse(sparse_uri):
    with tiledb.open(sparse_uri) as a:
        # 3 cells/row, 3*4 bytes for dims + 8+4+1=13 bytes for attrs = 25 bytes/cell
        assert estimate_row_bytes(a) == 75
        # 3 cells/row, 3*4 bytes for dims + 8+1=9 bytes for attrs = 21 bytes/cell
        assert estimate_row_bytes(a, attrs=["af8", "au1"]) == 63
        # 3 cells/row, 3*4 bytes for dims + 4 bytes for attrs = 16 bytes/cell
        assert estimate_row_bytes(a, attrs=["af4"]) == 48


@pytest.mark.parametrize("memory_budget", [2**i for i in range(14, 20)])
@pytest.mark.parametrize(
    "attrs",
    [(), ("af8",), ("af4",), ("au1",), ("af8", "af4"), ("af8", "au1"), ("af4", "au1")],
)
def test_get_max_buffer_size(dense_uri, memory_budget, attrs):
    config = {
        "sm.memory_budget": memory_budget,
        "py.max_incomplete_retries": 0,
    }
    with tiledb.scope_ctx(config), tiledb.open(dense_uri) as a:
        buffer_size = get_max_buffer_size(a.schema, attrs)
        # Check that the buffer size is a multiple of the row tile extent
        assert buffer_size % a.dim(0).tile == 0

        # Check that we can slice with buffer_size without incomplete reads
        query = a.query(attrs=attrs or None)
        for offset in range(0, a.shape[0], buffer_size):
            query[offset : offset + buffer_size]

        if buffer_size < a.shape[0]:
            # Check that buffer_size is the max size we can slice without incomplete reads
            buffer_size += 1
            with pytest.raises(tiledb.TileDBError):
                for offset in range(0, a.shape[0], buffer_size):
                    query[offset : offset + buffer_size]
